/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                          Disk.c                         **/
/**                                                         **/
/** This file contains standard disk access drivers working **/
/** with disk images from files.                            **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994-2001                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

#include "dream.h"
#include "MSX.h"

static char DriveData[737280];
static int Drives[2] = { -1,-1 };        /* Disk image files */
static int RdOnly[2];                    /* 1 = read-only    */ 

/** DiskPresent() ********************************************/
/** Return 1 if disk drive with a given ID is present.      **/
/*************************************************************/
byte DiskPresent(byte ID)
{ return((ID<MAXDRIVES)&&(Drives[ID]>=0)); }

/** DiskRead() ***********************************************/
/** Read requested sector from the drive into a buffer.     **/
/*************************************************************/
byte DiskRead(byte ID,byte *Buf,int N)
{
  int i;

  if((ID<MAXDRIVES)&&(Drives[ID]>=0))
  {
    for (i = N * 512; i < ((N+1) * 512); i++)
      Buf[i - (N * 512)] = DriveData[i];
    return (1);
  }
  else 
    return(0);
}

/** DiskWrite() **********************************************/
/** Write contents of the buffer into a given sector of the **/
/** disk.                                                   **/
/*************************************************************/
byte DiskWrite(byte ID,byte *Buf,int N)
{
/*  if((ID<MAXDRIVES)&&(Drives[ID]>=0)&&!RdOnly[ID])
    if(lseek(Drives[ID],N*512L,0)==N*512L)
      return(write(Drives[ID],Buf,512)==512);
*/
  return(0);
}

/** ChangeDisk() *********************************************/
/** Change disk image in a given drive. Closes current disk **/
/** image if Name=0 was given. Returns 1 on success or 0 on **/
/** failure.                                                **/
/*************************************************************/
byte ChangeDisk(byte ID,char *Name)
{
  int amt, fd;

  /* We only have MAXDRIVES drives */
  if(ID>=MAXDRIVES) return(0);
  /* Close previous disk image */
  if(Drives[ID]>=0) { Drives[ID]=-1; }
  /* If no disk image given, consider drive empty */
  if(!Name) return(1);
  /* Open new disk image */
  fd = cd_open(Name, O_RDONLY);
	if (fd < 0) {
		printf("Can't open Disk 1 for reading.\r\n");
		return 0;
	}
	
	amt = cd_read(fd, DriveData, 737280);
	cd_close(fd);

  Drives[ID] = 1;
  RdOnly[ID]=1;

  /* Return operation result */
  return(Drives[ID]>=0);
}

