/** ultraMSX: N64 MSX emulator *******************************/
/**                                                         **/
/**                           N64.c                         **/
/**       Dreamcast specific code by Jos Kwanten 2001		    **/
/** Copyright (C) Marat Fayzullin 1994,1995,1996            **/
/**               Elan Feingold   1995                      **/
/**               Ville Hallik    1996                      **/
/**															                            **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

/** Standard Playstation #includes ********************************/
#include "types.h"
#include "ay8910.h"
#include "MSX.h"
#include "scc.h"
#include "ay8910.h"
#include "dream.h"
#include "mem.h"
 
#define WIDTH  256
#define HEIGHT 212
#define VIDEO_BASE   0xA0300000
#define SND_BUFF_SIZE  704
#define SND_SAMPLE_FREQ 22050
#define THOLD 40

#define u_long	unsigned int

short Buf1[SND_BUFF_SIZE];	/* Soundbuffers for playing PSG*/
short Buf2[SND_BUFF_SIZE];	/* Soundbuffers for playing PSG*/
short Buf3[SND_BUFF_SIZE];	/* Soundbuffers for playing PSG*/
short Mix1[SND_BUFF_SIZE];	/* Soundbuffers for playing PSG / SCC*/
short Mix2[SND_BUFF_SIZE];	/* Soundbuffers for playing PSG / SCC*/
int   SwapBuffer=0;	/* Swap buffer N64 sound dma is dubble buffered  */
u8	 *SBuf, *XBuf;
u32  *VideoBuf;
int  XPal[16];		// 32bit Color mode
byte Pal[16][3];
long ColBuf[16*256];
int ColCnt=0;
int LTrigDown, RTrigDown;
volatile unsigned long *pos = (unsigned long*)0xa080ffc4;

int16 keystorest[10];

int curbuffer = 0;

#ifdef BPP24
u32 MSXPal[] = 
{
        0x00000000, /* Transparent */
        0x00000000, /* Black */
        0x0020C020, /* Medium Green */
        0x0060E060, /* Light Green */
        0x002020E0, /* Dark Blue */
        0x004060E0, /* Light Blue */
        0x00A02020, /* Dark Red */
        0x0040C0E0, /* Cyan */
        0x00E02020, /* Medium Red */
        0x00E06060, /* Light Red */
        0x00C0C020, /* Dark Yellow */
        0x00C0C080, /* Light Yellow */
        0x00208020, /* Dark Green */
        0x00C040A0, /* Magenta */
        0x00A0A0A0, /* Gray */
        0x00E0E0E0  /* White */
};
#else
unsigned short MSXPal[] = 
{
        0x0000, /* Transparent */
        0x0000, /* Black */
        0x1304, /* Medium Green */
        0x338c, /* Light Green */
        0x109c, /* Dark Blue */
        0x219c, /* Light Blue */
        0x5084, /* Dark Red */
        0x231c, /* Cyan */
        0x7084, /* Medium Red */
        0x718c, /* Light Red */
        0x6304, /* Dark Yellow */
        0x6310, /* Light Yellow */
        0x1204, /* Dark Green */
        0x6114, /* Magenta */
        0x5294, /* Gray */
        0x739c  /* White */
};
#endif

/** Keyboard bindings ****************************************/
byte Keys[256][2] =
{
/* 00 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{2,0x40},{2,0x80},{3,0x01},{3,0x02},
  {3,0x04},{3,0x08},{3,0x10},{3,0x20},{3,0x40},{3,0x80},{4,0x01},{4,0x02},
/* 10 */
  {4,0x04},{4,0x08},{4,0x10},{4,0x20},{4,0x40},{4,0x80},{5,0x01},{5,0x02},
  {5,0x04},{5,0x08},{5,0x10},{5,0x20},{5,0x40},{5,0x80},{0,0x02},{0,0x04},
/* 20 */
  {0,0x08},{0,0x10},{0,0x20},{0,0x40},{0,0x80},{1,0x01},{1,0x02},{0,0x01},
  {7,0x80},{7,0x04},{7,0x20},{7,0x08},{8,0x01},{1,0x04},{1,0x08},{2,0x02},
/* 30 */
	{2,0x20},{1,0x10},{1,0x20},{1,0x80},{1,0x10},{1,0x10},{2,0x04},{2,0x08},
  {2,0x10},{6,0x08},{6,0x20},{6,0x40},{6,0x80},{7,0x01},{7,0x02},{0,0x00},
/* 40 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{8,0x04},{8,0x02},{7,0x40},{8,0x08},{6,0x10},{7,0x10},{8,0x80},
/* 50 */
  {8,0x10},{8,0x40},{8,0x20},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 60 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 70 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 80 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{1,0x40},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 90 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* A0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* B0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* C0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* D0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* E0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* F0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{6,0x01},{6,0x02},{6,0x04} // Used for SHIFT, CTRL and ALT
};



// low-level pad buffers: never need to touch
unsigned char *bb0, *bb1;

/** PUTIMAGE: macro copying image buffer into a window ********/
void PutImage(void)
{
  int itx, ity, height;
	u32 offset1, offset2;
#ifdef BPP24
  u32 *vramptr;
  vramptr = vram_l;
#else
  u16 *vramptr;
  vramptr = vram_s;
#endif

// 	vid_waitvbl();
  
  height = HEIGHT;
  
  if(ScrMode < 4)
    {  
      for (ity = HEIGHT - 10 ; ity < height + 10 ; ity++) 
        {
          offset1 = ity*320;
          //offset2 = WIDTH*(ity-10);
          for (itx = 32; itx < WIDTH + 32; itx+=8) 
            {   
              vramptr[offset1+itx   ]= 0 ;
	            vramptr[offset1+itx +1]= 0;
	            vramptr[offset1+itx +2]= 0;
	            vramptr[offset1+itx +3]= 0;
	            vramptr[offset1+itx +4]= 0;
	            vramptr[offset1+itx +5]= 0;
	            vramptr[offset1+itx +6]= 0;
	            vramptr[offset1+itx +7]= 0;
            }
        }
      height = HEIGHT - 20;
    }
  
  for (ity = 10 ; ity < height + 10 ; ity++) 
    {
      offset1 = ity*320;
      offset2 = WIDTH*(ity-10);
      for (itx = 32; itx < WIDTH + 32; itx+=8) 
        {   
          vramptr[offset1+itx]=MSXPal[XBuf[offset2+itx-32]];
	        vramptr[offset1+itx +1]=MSXPal[XBuf[offset2+itx-32+1]];
	        vramptr[offset1+itx +2]=MSXPal[XBuf[offset2+itx-32+2]];
	        vramptr[offset1+itx +3]=MSXPal[XBuf[offset2+itx-32+3]];
	        vramptr[offset1+itx +4]=MSXPal[XBuf[offset2+itx-32+4]];
	        vramptr[offset1+itx +5]=MSXPal[XBuf[offset2+itx-32+5]];
	        vramptr[offset1+itx +6]=MSXPal[XBuf[offset2+itx-32+6]];
	        vramptr[offset1+itx +7]=MSXPal[XBuf[offset2+itx-32+7]];
        }
    }
}

int ReadControllers(void)
{
    return(0);
}

/** Screen Mode Handlers [number of screens + 1] *************/
struct
{                                  /* Screen mode handlers:  */
  int (*Init)();                   /* Initialize screen      */
  void (*Colors)();                /* Update colors/palette  */
  void (*Refresh)(byte Y1,byte Y2);/* Refresh screen         */
} SCR[MAXSCREEN+2] =               /* Number of screens + 1  */
{
  { InitScrF,ColorsScrF,RefreshScr0 }, /* SCR 0:TEXT 40x24  */
  { InitScrF,ColorsScrF,RefreshScr1 }, /* SCR 1:TEXT 32x24  */
  { InitScrF,ColorsScrF,RefreshScr2 }, /* SCR 2:BLK 256x192 */
  { InitScrF,ColorsScrF,RefreshScr3 }, /* SCR 3:64x48x16    */
  { InitScrF,ColorsScrF,RefreshScr4 }, /* SCR 4:BLK 256x192 */
  { InitScrF,ColorsScrF,RefreshScr5 }, /* SCR 5:256x192x16  */
  { InitScrF,ColorsScrF,RefreshScr6 }, /* SCR 6:512x192x4   */
  { InitScrF,ColorsScrF,RefreshScr7 }, /* SCR 7:512x192x16  */
  { InitScrF,ColorsScrF,RefreshScr8 }, /* SCR 8:256x192x256 */
  { InitScrF,ColorsScrF,RefreshTx80 }  /* SCR 0:TEXT 80x24  */
};

/** TrashMachine *********************************************/
/** Deallocate all resources taken by InitMachine().        **/
/*************************************************************/
void TrashMachine(void)
{
  free(XBuf);
  free(SBuf);
  if(Verbose) printf("Shutting down...\n");
}


/** InitMachine **********************************************/
/** Allocate resources needed by Playstation dependent code.**/
/*************************************************************/
int InitMachine(void)
{
	int i;
  if(Verbose) printf("OK\n", 0);
  // memory allocation for buffers..
  SBuf=(u8 *) malloc(4*256*212);
  XBuf=(u8 *) malloc(4*WIDTH*HEIGHT);
  for(i = 0; i < (4 * WIDTH * HEIGHT); i++)
    XBuf[i] = 0;
  return(1);
}

/** Keyboard *************************************************/
/** Check for keyboard events, parse them, and modify MSX   **/
/** keyboard matrix.                                        **/
/*************************************************************/
void Keyboard(void)
{
	uint8 mcont;
	uint8 kbcont;
	uint8 counter;
	cont_cond_t cond;
	kbd_cond_t kbcond;

	mcont = maple_controller_addr();
	if (!mcont) {
		printf("No controllers attached\r\n");
	}
	else if (cont_get_cond(mcont, &cond)) 
  {
	/* Check for start on the controller */
		  printf("Error getting controller status\r\n");
	}
	else 
	{
	// Stick functions
		if(!(cond.buttons  & CONT_DPAD_UP)||(cond.joyy < 128 - THOLD)){
			// pressed is 0
			KeyMap[8]&=~0x20;
			//while(1);
		}else{
			// depressed is 1
			KeyMap[8]|=0x20;
		}
  
		if(!(cond.buttons  & CONT_DPAD_DOWN)||(cond.joyy > 128 + THOLD)){
			// pressed is 0
			KeyMap[8]&=~0x40;
		}else{
			// depressed is 1
			KeyMap[8]|=0x40;
		}
  
		if(!(cond.buttons  & CONT_DPAD_LEFT)||(cond.joyx < 128 - THOLD)){
			// pressed is 0
			KeyMap[8]&=~0x10;
		}else{
			// depressed is 1
			KeyMap[8]|=0x10;
		}
  
		if(!(cond.buttons  & CONT_DPAD_RIGHT)||(cond.joyx > 128 + THOLD)){
			// pressed is 0
			KeyMap[8]&=~0x80; //0x80
		}else{
			// depressed is 1
			KeyMap[8]|=0x80;
		}

		if(!(cond.buttons  & CONT_A)){
			// pressed is 0
			KeyMap[8]&=~0x01;
		}else{
			// depressed is 1
			KeyMap[8]|=0x01;
		}

		if(!(cond.buttons  & CONT_B)){
			// pressed is 0
			KeyMap[4]&=~0x04;
		}else{
			// depressed is 1
			KeyMap[4]|=0x04;
		}
  
  
		if(!(cond.buttons  & CONT_START)){
			// pressed is 0
			KeyMap[6]&=~0x20;
		}else{
			// depressed is 1
			KeyMap[6]|=0x20;
		}
  
		if(!(cond.buttons  & CONT_X)){
			KeyMap[6]&=~0x01;
		}
		else
		{
			KeyMap[6]|=0x01;
		}

		if(!(cond.buttons  & CONT_Y)){
			KeyMap[7]&=~0x02;
		}
		else
		{
			KeyMap[7]|=0x02;
		}

/*
		if(rdata[0].button & L_CBUTTONS){
			// pressed is 0
			KeyMap[6]&=~0x40;
		}else{
			// depressed is 1
			KeyMap[6]|=0x40;
		}

		if(rdata[0].button & U_CBUTTONS){
			// pressed is 0
			KeyMap[6]&=~0x80;
		}else{
			// depressed is 1
			KeyMap[6]|=0x80;
		}
  
		if(rdata[0].button & R_CBUTTONS){
			// pressed is 0
			KeyMap[7]&=~0x01;
		}else{
			// depressed is 1
			KeyMap[7]|=0x01;
		}
  
		if(rdata[0].button & D_CBUTTONS){
			// pressed is 0
			KeyMap[7]&=~0x02;
		}else{
			// depressed is 1
			KeyMap[7]|=0x02;
		}
*/  
		if(cond.ltrig){
			printf("Exit....\r\n");
			CPURunning = 0;
//  			if((LTrigDown==0)&&(UPeriod >1))
//  				UPeriod--;		  
//  			LTrigDown = 1;
		}
//		else
//				LTrigDown = 0;
/*  
		if(rdata[0].button & R_TRIG){
  			if(RTrigDown==0)  	  
  	  			UPeriod++;
  			RTrigDown = 1;
		}
		else
			RTrigDown = 0;
*/	
	}
/* Second controller */
/*
	mcont = maple_controller_addr_num(1);
	if (mcont && !cont_get_cond(mcont, &cond)) 
	{
		// Stick functions
		if(!(cond.buttons  & CONT_DPAD_UP)||(cond.joyy < 128 - THOLD)){
			// pressed is 0
			KeyMap[8]&=~0x20;
			//while(1);
		}else{
			// depressed is 1
			KeyMap[8]|=0x20;
		}
  
		if(!(cond.buttons  & CONT_DPAD_DOWN)||(cond.joyy > 128 + THOLD)){
			// pressed is 0
			KeyMap[8]&=~0x40;
		}else{
			// depressed is 1
			KeyMap[8]|=0x40;
		}
  
		if(!(cond.buttons  & CONT_DPAD_LEFT)||(cond.joyx < 128 - THOLD)){
			// pressed is 0
			KeyMap[8]&=~0x10;
		}else{
			// depressed is 1
			KeyMap[8]|=0x10;
		}
  
		if(!(cond.buttons  & CONT_DPAD_RIGHT)||(cond.joyx > 128 + THOLD)){
			// pressed is 0
			KeyMap[8]&=~0x80; //0x80
		}else{
			// depressed is 1
			KeyMap[8]|=0x80;
		}

		if(!(cond.buttons  & CONT_A)){
			// pressed is 0
			KeyMap[8]&=~0x01;
		}else{
			// depressed is 1
			KeyMap[8]|=0x01;
		}

		if(!(cond.buttons  & CONT_B)){
			// pressed is 0
			KeyMap[4]&=~0x04;
		}else{
			// depressed is 1
			KeyMap[4]|=0x04;
		}
  
  
		if(!(cond.buttons  & CONT_START)){
			// pressed is 0
			KeyMap[6]&=~0x20;
		}else{
			// depressed is 1
			KeyMap[6]|=0x20;
		}
	}
*/

	kbcont = maple_kb_addr();
	if (kbcont && !kbd_get_cond(kbcont, &kbcond))
	{
		for (counter = 0; counter < 9; counter++)
		{
			if (keystorest[counter] >= 0)
				KeyMap[Keys[keystorest[counter]][0]]|=Keys[keystorest[counter]][1];
		}
		for (counter = 0; counter < 6; counter++)
		{
			if (kbcond.keys[counter] > 1)
			{
//				printf("Key code: %d\n",kbcond.keys[counter]);
	      KeyMap[Keys[kbcond.keys[counter]][0]]&=~Keys[kbcond.keys[counter]][1];
				keystorest[counter] = kbcond.keys[counter];
			}
			else
				keystorest[counter] = -1;
		}
		if ((kbcond.modifiers & KBD_MOD_LSHIFT)
			||(kbcond.modifiers & KBD_MOD_RSHIFT))
		{
			KeyMap[6]&=~0x01;
			keystorest[6] = 0xfd;
		}
		else
		{
			keystorest[6] = -1;
		}
		if ((kbcond.modifiers & KBD_MOD_LALT)
			||(kbcond.modifiers & KBD_MOD_RALT))
		{
			KeyMap[6]&=~0x04;
			keystorest[7] = 0xff;
		}
		else
		{
			keystorest[7] = -1;
		}
		if ((kbcond.modifiers & KBD_MOD_LCTRL)
			||(kbcond.modifiers & KBD_MOD_RCTRL))
		{
			KeyMap[6]&=~0x02;
			keystorest[8]= 0xfe;
		}
		else
		{
			keystorest[8] = -1;
		}
	}

	//printf("X = %d, Y = %d", cond.joyx, cond.joyy);
		
}

/** InitScrF *************************************************/
/** Generic function to be called when screen is changed.   **/
/*************************************************************/
int InitScrF(void)
{
  if(Verbose&0x02) printf("Screen %d initialized\n",ScrMode);
  return(1);
}


/** ColorScrF ************************************************/
/** Generic function to be called when colors are changed.  **/
/*************************************************************/
void ColorsScrF(void)
{
  	byte J;
  	for(J=0; J<16; J++)
  	{
/*  		MSXPal[J] =  Palette[J][0] * 65536 +
    	             Palette[J][1] * 256 +
    	             Palette[J][2];
*/
			#ifdef BPP24
  				MSXPal[J] = (Palette[J][0] << 16) +
    									(Palette[J][1] << 8) +
    									Palette[J][2];
			#else
  				MSXPal[J] = ((Palette[J][0]>>3) << 10) +
    									((Palette[J][1]>>3) << 5) +
    									(Palette[J][2]>>3);
			#endif    
		}
}


/** Part of the code common for Unix/X and MSDOS drivers *****/ 
#include "Common.h"


/** RefreshScr6 **********************************************/
/** Function to be called to update SCREEN 6.               **/
/*************************************************************/
void RefreshScr6(register byte Y1,register byte Y2)
{
  register byte X,J,C;
  register byte *P,*T,*R;

  P=XBuf;//+WIDTH*(HEIGHT-192)/2+Y1*WIDTH;

  if(ScreenON)
  {
    ColorSprites();

    R=SBuf+((word)Y1<<8);
    P+=(WIDTH-256)/2;
    J=VDP[23]+Y1;
    T=ChrTab+((word)J<<7);

    for(;Y1<=Y2;Y1++,J++,P+=WIDTH-256)
    {
      if(!J) T=ChrTab;
      for(X=0;X<32;X++)
      {
        C=R[0];C=((C>>4)|C)&0x0F;P[0]=C? C:T[0]>>6;
        C=R[1];C=((C>>4)|C)&0x0F;P[1]=C? C:(T[0]>>2)&0x03;
        C=R[2];C=((C>>4)|C)&0x0F;P[2]=C? C:T[1]>>6;
        C=R[3];C=((C>>4)|C)&0x0F;P[3]=C? C:(T[1]>>2)&0x03;
        C=R[4];C=((C>>4)|C)&0x0F;P[4]=C? C:T[2]>>6;
        C=R[5];C=((C>>4)|C)&0x0F;P[5]=C? C:(T[2]>>2)&0x03;
        C=R[6];C=((C>>4)|C)&0x0F;P[6]=C? C:T[3]>>6;
        C=R[7];C=((C>>4)|C)&0x0F;P[7]=C? C:(T[3]>>2)&0x03;
        R+=8;P+=8;T+=4;
      }
    }  
  }
  else memset(P,XPal[BGColor],(Y2-Y1+1)*WIDTH);

  if(EndOfFrame) PutImage();
} 

/** RefreshScr7 **********************************************/
/** Function to be called to update SCREEN 7.               **/
/*************************************************************/
void RefreshScr7(register byte Y1,register byte Y2)
{
  if(EndOfFrame) PutImage();
}
    
   
/** RefreshScr8 **********************************************/
/** Function to be called to update SCREEN 8.               **/
/*************************************************************/
void RefreshScr8(register byte Y1,register byte Y2)
{
  if(EndOfFrame) PutImage();
}


/** RefreshTx80 **********************************************/
/** Function to be called to update TEXT80.                 **/
/*************************************************************/
void RefreshTx80(register byte Y1,register byte Y2)
{
  if(EndOfFrame) PutImage();
}


/** InitSound ************************************************/
/** Function to be called to init N64 sound dma		        **/
/*************************************************************/
void InitSound(void)
{
  AY8910_sh_start(SND_SAMPLE_FREQ);
  SCC_sh_start(SND_SAMPLE_FREQ);  
}


/** DoSound	**************************************************/
/** Function to be called to update N64 soundbuffers        **/
/*************************************************************/
void DoSound(void)
{
 int i;
 int val, realbuffer;
 
 for(i=0; i < SND_BUFF_SIZE; i++) Mix1[i] = 0;
 
 val = *pos;
 snd_ram_write_wait();
 
 realbuffer = !(val < 704);
      
 if (curbuffer != realbuffer) 
   {
	    /* Yep, adjust "current" buffer and initiate a load */
      AY8910Update(0, Buf1, Buf2, Buf3, SND_BUFF_SIZE);
      if(UseSCC)
        {	
          SCCUpdate8(1, Mix1, SND_BUFF_SIZE);
          for(i=0; i < SND_BUFF_SIZE; i++)
              Mix1[i] = Mix1[i]/2 + ((Buf1[i]+Buf2[i]+Buf3[i]))/5;
        }
      else 
        {
          for(i=0; i < SND_BUFF_SIZE; i++)
          Mix1[i] = ((Buf1[i]+Buf2[i]+Buf3[i]))/5;
        }
   
      snd_load(Mix1, 0x10000 + 1408*curbuffer, 1407);
      //printf("Loaded %d\r\n", curbuffer);
      curbuffer = realbuffer;
   }
}
