/*
** Sound.c
**
** This file countains PSG/SCC emulation stuff for fMSX via the /dev/dsp
** device present in Linux and FreeBSD (and probably in other PC Unices
** using USS(/Lite) (formerly known as VoxWare) sound driver). Also very
** experimental support for Sun's /dev/audio is present.
**
** (C) Ville Hallik (ville@physic.ut.ee) 1996
**
** You can do almost anything you want with this file, provided that notice
** about original author still appears there and in case you make any
** modifications, PLEASE label it as modified version.
**
** Known problems:
** 1. The noise and envelope frequencies may be wrong - I dont have
**    real MSX handy to compare them.
** 2. The seventh bit of port 0xAA is not supported (it's almost impossible
**    to do with the method I use here), fortunately it's not used in most
**    cases.
** 3. Modifications of PSG registers are checked (theoretically ;) once
**    per every SOUND_BUFSIZE/SOUND_RATE seconds. It's about 11ms with the
**    values of 256bytes and 22050Hz, which should be satisfactory for most
**    cases.
** 4. Support for Sun's /dev/audio is crap, but it works somehow ;)
** 5. I don't know how realistic is the emulation of SCC chip.
**
** last modified on Sat Apr 22 1996
*/

//#define SOUND

#ifdef SOUND

/* We don't need this on an N64
#include "MSX.h"
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <linux/soundcard.h>*/

#define AUDIO_CONV(A) (128+(A))

/*
** Output rate (Hz)
*/
#ifndef SOUND_RATE
#define SOUND_RATE 22050
#endif SOUND_RATE

/*
** Buffer size. SOUND_BUFSIZE should be equal to (or at least less than)
** 2 ^ SOUND_BUFSIZE_BITS
*/
#define SOUND_BUFSIZE 256
#define SOUND_BUFSIZE_BITS 8

/*
** Number of audio buffers to use. Must be at least 2, but bigger value
** results in better behaviour on loaded systems (but output gets more
** delayed)
*/
#define SOUND_NUM_OF_BUFS 8


//pid_t soundchildpid=0;
//int soundpipe[2];
//int sounddev=-1;
int sound_rate=SOUND_RATE;
int UseSound=1;
int UseSCC=1;
u8 SoundBuff[SOUND_NUM_OF_BUFS][SOUND_BUFSIZE];


unsigned char envforms[16][32] = {
	{ 15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{ 15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{ 15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{ 15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{ 15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0,
	  15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0 },
	{ 15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{ 15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0,
	   0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15 },
	{ 15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0,
	  15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15 },
	{  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	   0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15 },
	{  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	  15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15 },
	{  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	  15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0 },
	{  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 }
};

void SoundMainLoop(void);

void TrashSound(void)
{
  //int J;
  //if(soundchildpid) { kill(soundchildpid,SIGKILL);wait(&J); }
  //if(sounddev!=-1) close(sounddev);
}

int OpenSoundDev(void)
{
  	int I,J,K;
    /* Setup sound */
  	IO_WRITE(AI_DACRATE_REG, VI_PAL_CLOCK / SOUND_RATE);
  	IO_WRITE(AI_BITRATE_REG, SOUND_BUFSIZE_BITS - 1);
   	//IO_WRITE(AI_DRAM_ADDR_REG, Buffer1);
   	//IO_WRITE(AI_LEN_REG, 22048);
   	//IO_WRITE(AI_CONTROL_REG, AI_CONTROL_DMA_ON);

  return(1);
}

int InitSound(void )
{

  if(!UseSound) return;

  if(Verbose) puts("Starting sound server:");

  UseSound=0;sounddev=-1;soundchildpid=0;

  if( !OpenSoundDev() ) return(0);

  UseSound=1;return(1);
}

void SoundOut(byte R,byte V)
{
  static unsigned char Buf[1024];
  static int Buffered = 0;

  if(UseSound) {
    if((R==0xFF)||(Buffered>=sizeof(Buf))) {
      write(soundpipe[1],Buf,Buffered);
      Buffered = 0;
    }
    if(R!=0xFF) {
      Buf[Buffered++]=R; Buf[Buffered++]=V;
    }
  }
}

void FlushSound(void)      { SoundOut(0xFF,0); }
void StopSound(void)       { if(UseSound) kill(soundchildpid,SIGUSR1); }
void ResumeSound(void)     { if(UseSound) kill(soundchildpid,SIGUSR2); }
void PSGOut(byte R,byte V) { if(R<16||R==0xFF) SoundOut(R,V); }
void SCCOut(byte R,byte V) { if(UseSCC&&R<0x90) SoundOut(R+0x10,V); }

void SoundSignal( int sig )
{
  static ok_to_continue=0;
  
  switch( sig ) {
  case SIGUSR1:
    /* Suspend execution, until SIGUSR2 catched */
#ifndef SUN_AUDIO
    ioctl( sounddev, SNDCTL_DSP_RESET );
#endif /* not SUN_AUDIO */
    close( sounddev );
    ok_to_continue=0;
    while( !ok_to_continue ) pause();
    if( Verbose ) puts( "Soundserver: Re-opening sound device:" );
    OpenSoundDev();
    if( Verbose ) fflush( stdout );
    signal( SIGUSR1, SoundSignal );
    break;
  case SIGUSR2:
    ok_to_continue=1;
    signal( SIGUSR2, SoundSignal );
    break;
  }
}

void SoundMainLoop( void )
{
  int incr0=0, incr1=0, incr2=0, increnv=0, incrnoise=0;
  int statenoise=0, noisegen=1;
  int counter0, counter1, counter2, countenv, countnoise;
  int r=0, v=0, i, j=0, rc, x;
  int vol0, vol1, vol2, volnoise, envelope = 15;
  int c0, c1, l0, l1, l2;
  unsigned char buf[2];
  unsigned char soundbuf[SOUND_BUFSIZE];
  unsigned char PSG[16];
  pid_t parent;
  /* Variables for SCC emulation: */
  int SCCactive=0, SCCoutv=0;
  int SCCinc1=0, SCCinc2=0, SCCinc3=0, SCCinc4=0, SCCinc5=0;
  int SCCcnt1=0, SCCcnt2=0, SCCcnt3=0, SCCcnt4=0, SCCcnt5=0;
  int SCCvol1, SCCvol2, SCCvol3, SCCvol4, SCCvol5;
  unsigned char SCC[256];
	
  parent = getppid();
  signal( SIGUSR1, SoundSignal );
  signal( SIGUSR2, SoundSignal );
	
  for( i=0; i<16; ++i ) PSG[i] = 0;
  PSG[7] = 077;
  PSG[8] = 8;
  PSG[9] = 8;
  PSG[10] = 8;
  for( i=0; i<256; ++i ) SCC[i] = 0;
	
  fcntl( soundpipe[0], F_SETFL, O_NONBLOCK );
	
  for( ;; ) {
    /* check for parent's existence */
    if( parent != getppid() ) {
      fprintf( stderr, "Parent died\n" );
      exit( 1 );
    }
    /* Read from pipe: */
    for( ;; ) {
      rc = read( soundpipe[0], buf, v < 0 ? 1 : 2 );
      if( rc == 0 ) break;
      if( rc < 0 ) {
        if( errno == EWOULDBLOCK ) {
          break;
        } else {
          perror( "read" );
          exit( 1 );
        }
      }
      if( v == -1 ) {
        v = buf[0];
      } else {
        r = buf[0];
        if( rc == 1 ) {
          v = -1;
          break;
        }
        v = buf[1];
      }
			
      if( r<16 ) {
      
        PSG[r] = v;
			
        switch( r ) {
        case 0:
        case 1:
          x = (PSG[0]+((unsigned)(PSG[1]&0xF)<<8));
          incr0 = x ? 1832727040 / sound_rate * 4 / x : 0;
          break;
        case 2:
        case 3:
          x = (PSG[2]+((unsigned)(PSG[3]&0xF)<<8));
          incr1 = x ? 1832727040 / sound_rate * 4 / x : 0;
          break;
        case 4:
        case 5:
          x = (PSG[4]+((unsigned)(PSG[5]&0xF)<<8));
          incr2 = x ? 1832727040 / sound_rate * 4 / x : 0;
          break;
        case 6:
          x = PSG[6]&0x1F;
          incrnoise = 1832727040 / sound_rate * 4 / ( x ? x : 1 );
          break;
        case 8:
          PSG[8] &= 0x1F;
          break;
        case 9:
          PSG[9] &= 0x1F;
          break;
        case 10:
          PSG[10] &= 0x1F;
          break;
        case 11:
        case 12:
        case 13:
          PSG[13] &= 0xF;
          x = (PSG[11]+((unsigned)PSG[12]<<8));
          increnv = x ? 1832727040 / sound_rate * 4 / x * SOUND_BUFSIZE : 0;
          countenv=0;
          break;
        }

      } else if( UseSCC ) {

        if( !SCCactive ) {
          if( Verbose ) {
            puts( "Soundserver: SCC emulation activated" );
            fflush( stdout );
          }
          SCCactive=1;
        }
        SCC[r-0x10]=v;

        switch(r-0x10) {
        case 0x80:
        case 0x81:
          x = (SCC[0x80]+((unsigned)(SCC[0x81]&0xF)<<8));
          SCCinc1 = x ? 1832727040 / sound_rate * 4 / x : 0;
          break;
        case 0x82:
        case 0x83:
          x = (SCC[0x82]+((unsigned)(SCC[0x83]&0xF)<<8));
          SCCinc2 = x ? 1832727040 / sound_rate * 4 / x : 0;
          break;
        case 0x84:
        case 0x85:
          x = (SCC[0x84]+((unsigned)(SCC[0x85]&0xF)<<8));
          SCCinc3 = x ? 1832727040 / sound_rate * 4 / x : 0;
          break;
        case 0x86:
        case 0x87:
          x = (SCC[0x86]+((unsigned)(SCC[0x87]&0xF)<<8));
          SCCinc4 = x ? 1832727040 / sound_rate * 4 / x : 0;
          break;
        case 0x88:
        case 0x89:
          x = (SCC[0x88]+((unsigned)(SCC[0x89]&0xF)<<8));
          SCCinc5 = x ? 1832727040 / sound_rate * 4 / x : 0;
          break;
        }

      }
      
    }

    envelope = envforms[PSG[13]][(countenv>>16)&0x1F];
    if( ( countenv += increnv ) & 0xFFE00000 ) {
      switch( PSG[13] ) {
      case 8:
      case 10:
      case 12:
      case 14:
        countenv -= 0x200000;
        break;
      default:
        countenv = 0x100000;
        increnv = 0;
      }
    }

    vol0 = ( PSG[8] < 16 ) ? PSG[8] : envelope;
    vol1 = ( PSG[9] < 16 ) ? PSG[9] : envelope;
    vol2 = ( PSG[10] < 16 ) ? PSG[10] : envelope;

    volnoise = (
     ( ( PSG[7] & 010 ) ? 0 : vol0 ) +
     ( ( PSG[7] & 020 ) ? 0 : vol1 ) +
     ( ( PSG[7] & 040 ) ? 0 : vol2 ) ) / 2;

    vol0 = ( PSG[7] & 001 ) ? 0 : vol0;
    vol1 = ( PSG[7] & 002 ) ? 0 : vol1;
    vol2 = ( PSG[7] & 004 ) ? 0 : vol2;
		
    if( SCCactive ) {
      SCCvol1 = SCC[0x8F] & 0x01 ? SCC[0x8A] & 0xF : 0;
      SCCvol2 = SCC[0x8F] & 0x02 ? SCC[0x8B] & 0xF : 0;
      SCCvol3 = SCC[0x8F] & 0x04 ? SCC[0x8C] & 0xF : 0;
      SCCvol4 = SCC[0x8F] & 0x08 ? SCC[0x8D] & 0xF : 0;
      SCCvol5 = SCC[0x8F] & 0x10 ? SCC[0x8E] & 0xF : 0;
    }

    for( i=0; i<SOUND_BUFSIZE; ++i ) {

      if( SCCactive ) {

        SCCcnt1 = ( SCCcnt1 + SCCinc1 ) & 0xFFFF;
        c0 = (signed char)SCC[SCCcnt1>>11];
        c1 = (signed char)SCC[((SCCcnt1>>11)+1)&0x1F];
        SCCoutv = SCCvol1*(c0+(c1-c0)*(SCCcnt1&0x7FF)/0x800);

        SCCcnt2 = ( SCCcnt2 + SCCinc2 ) & 0xFFFF;
        c0 = (signed char)SCC[(SCCcnt2>>11)+0x20];
        c1 = (signed char)SCC[(((SCCcnt2>>11)+1)&0x1F)+0x20];
        SCCoutv += SCCvol2*(c0+(c1-c0)*(SCCcnt2&0x7FF)/0x800);

        SCCcnt3 = ( SCCcnt3 + SCCinc3 ) & 0xFFFF;
        c0 = (signed char)SCC[(SCCcnt3>>11)+0x40];
        c1 = (signed char)SCC[(((SCCcnt3>>11)+1)&0x1F)+0x40];
        SCCoutv += SCCvol3*(c0+(c1-c0)*(SCCcnt3&0x7FF)/0x800);

        SCCcnt4 = ( SCCcnt4 + SCCinc4 ) & 0xFFFF;
        c0 = (signed char)SCC[(SCCcnt4>>11)+0x60];
        c1 = (signed char)SCC[(((SCCcnt4>>11)+1)&0x1F)+0x60];
        SCCoutv += SCCvol4*(c0+(c1-c0)*(SCCcnt4&0x7FF)/0x800);

        SCCcnt5 = ( SCCcnt5 + SCCinc5 ) & 0xFFFF;
        c0 = (signed char)SCC[(SCCcnt5>>11)+0x60];
        c1 = (signed char)SCC[(((SCCcnt5>>11)+1)&0x1F)+0x60];
        SCCoutv += SCCvol5*(c0+(c1-c0)*(SCCcnt5&0x7FF)/0x800);

        SCCoutv /= 128;

      }

      /*
      ** These strange tricks are needed for getting rid
      ** of nasty interferences between sampling frequency
      ** and "rectangular sound" (which is also the output
      ** of real AY-3-8910) we produce.
      */

      c0 = counter0;
      c1 = counter0 + incr0;
      l0 = ((c0&0x8000)?-16:16);
      if( (c0^c1)&0x8000 ) {
        l0=l0*(0x8000-(c0&0x7FFF)-(c1&0x7FFF))/incr0;
      }
      counter0 = c1 & 0xFFFF;
			
      c0 = counter1;
      c1 = counter1 + incr1;
      l1 = ((c0&0x8000)?-16:16);
      if( (c0^c1)&0x8000 ) {
        l1=l1*(0x8000-(c0&0x7FFF)-(c1&0x7FFF))/incr1;
      }
      counter1 = c1 & 0xFFFF;
			
      c0 = counter2;
      c1 = counter2 + incr2;
      l2 = ((c0&0x8000)?-16:16);
      if( (c0^c1)&0x8000 ) {
        l2=l2*(0x8000-(c0&0x7FFF)-(c1&0x7FFF))/incr2;
      }
      counter2 = c1 & 0xFFFF;
			
      countnoise &= 0xFFFF;
      if( ( countnoise += incrnoise ) & 0xFFFF0000 ) {
        /*
        ** The following code is a random bit generator :)
        */
        statenoise =
         ( ( noisegen <<= 1 ) & 0x80000000
         ? noisegen ^= 0x00040001 : noisegen ) & 1;
      }

      soundbuf[i] = AUDIO_CONV ( SCCoutv +
       ( l0 * vol0 + l1 * vol1 + l2 * vol2 ) / 16 +
       ( statenoise ? volnoise : -volnoise ) );

    }

    if( sounddev == -1 ) {
      sleep(1);
      continue;
    }

    /*
    ** We'll block here until next DMA buffer becomes free.
    ** It happens once per (1<<SOUND_BUFSIZE_BITS)/sound_rate seconds.
    */
		
    write( sounddev, soundbuf, SOUND_BUFSIZE );

  }
}

#endif SOUND
