/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                           Unix.c                        **/
/**                                                         **/
/** This file contains Unix/X-dependent subroutines and     **/
/** drivers. It includes common drivers from Common.h.      **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994,1995,1996            **/
/**               Elan Feingold   1995                      **/
/**               Ville Hallik    1996                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

#define N64

#ifndef N64 
/** Standard Unix/X #includes ********************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#endif

#include "MSX.h"

  
/** MIT Shared Memory Extension for X ************************/
#ifdef MITSHM
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
XShmSegmentInfo SHMInfo;
int UseSHM=1;
#endif


/** Sound-related definitions ********************************/
#ifdef SOUND
int InitSound(void);
void TrashSound(void);
void FlushSound(void);
void StopSound(void);
void ResumeSound(void);
extern int UseSound;
#endif


/** Various X-related variables ******************************/
#ifndef N64
Display *Dsp;
Window Wnd;
Colormap DefaultCMap;
XImage *Img;
GC DefaultGC;
#endif
unsigned long White,Black;

/** Various variables and short functions ********************/
#define WIDTH  340
#define HEIGHT 240

char *Title="fMSX Unix/X 1.5";

int IFreq=50;
int SaveCPU=1;

//struct itimerval TimerValue;

byte *XBuf,*SBuf;
byte BPal[256],XPal[16],XPal0,Pal[16][3];
long ColBuf[16*256];
int ColCnt=0;

void OnBreak(int Arg) { CPURunning=0; }
void OnTimer(int Arg) { IFlag=1;//signal(SIGVTALRM,OnTimer); 
                       }


/** PUTIMAGE: macro copying image buffer into a window ********/
void PutImage(void)
{
#ifndef N64
#ifdef MITSHM
  if(UseSHM) XShmPutImage(Dsp,Wnd,DefaultGC,Img,0,0,0,0,WIDTH,HEIGHT,False);
  else
#endif
  XPutImage(Dsp,Wnd,DefaultGC,Img,0,0,0,0,WIDTH,HEIGHT);
  if(ColCnt) { XFreeColors(Dsp,DefaultCMap,ColBuf,ColCnt,0);ColCnt=0; }
  XFlush(Dsp);
#endif


}


/** Screen Mode Handlers [number of screens + 1] *************/
struct
{                                  /* Screen mode handlers:  */
  int (*Init)();                   /* Initialize screen      */
  void (*Colors)();                /* Update colors/palette  */
  void (*Refresh)(byte Y1,byte Y2);/* Refresh screen         */
} SCR[MAXSCREEN+2] =               /* Number of screens + 1  */
{
  { InitScrF,ColorsScrF,RefreshScr0 }, /* SCR 0:TEXT 40x24  */
  { InitScrF,ColorsScrF,RefreshScr1 }, /* SCR 1:TEXT 32x24  */
  { InitScrF,ColorsScrF,RefreshScr2 }, /* SCR 2:BLK 256x192 */
  { InitScrF,ColorsScrF,RefreshScr3 }, /* SCR 3:64x48x16    */
  { InitScrF,ColorsScrF,RefreshScr4 }, /* SCR 4:BLK 256x192 */
  { InitScrF,ColorsScrF,RefreshScr5 }, /* SCR 5:256x192x16  */
  { InitScrF,ColorsScrF,RefreshScr6 }, /* SCR 6:512x192x4   */
  { InitScrF,ColorsScrF,RefreshScr7 }, /* SCR 7:512x192x16  */
  { InitScrF,ColorsScrF,RefreshScr8 }, /* SCR 8:256x192x256 */
  { InitScrF,ColorsScrF,RefreshTx80 }  /* SCR 0:TEXT 80x24  */
};


/** Keyboard bindings ****************************************/
byte Keys[256][2] =
{
/* 00 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {7,0x20},{7,0x08},{0,0x00},{0,0x00},{0,0x00},{7,0x80},{0,0x00},{0,0x00},
/* 10 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{7,0x04},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 20 */
  {8,0x01},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{2,0x01},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{2,0x04},{1,0x04},{2,0x08},{2,0x10},
/* 30 */
  {0,0x01},{0,0x02},{0,0x04},{0,0x08},{0,0x10},{0,0x20},{0,0x40},{0,0x80},
  {1,0x01},{1,0x02},{0,0x00},{1,0x80},{0,0x00},{1,0x08},{0,0x00},{0,0x00},
/* 40 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 50 */
  {0,0x00},{8,0x10},{8,0x20},{8,0x80},{8,0x40},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{1,0x20},{1,0x10},{1,0x40},{0,0x00},{0,0x00},
/* 60 */
  {2,0x02},{2,0x40},{2,0x80},{3,0x01},{3,0x02},{3,0x04},{3,0x08},{3,0x10},
  {3,0x20},{3,0x40},{3,0x80},{4,0x01},{4,0x02},{4,0x04},{4,0x08},{4,0x10},
/* 70 */
  {4,0x20},{4,0x40},{4,0x80},{5,0x01},{5,0x02},{5,0x04},{5,0x08},{5,0x10},
  {5,0x20},{5,0x40},{5,0x80},{0,0x00},{0,0x00},{0,0x00},{6,0x04},{0,0x00},
/* 80 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 90 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* A0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* B0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{6,0x20},{6,0x40},
/* C0 */
  {6,0x80},{7,0x01},{7,0x02},{8,0x02},{8,0x08},{8,0x04},{7,0x40},{7,0x10},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* D0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {8,0x02},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* E0 */
  {0,0x00},{6,0x01},{6,0x08},{6,0x02},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{6,0x10},{6,0x04},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* F0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{8,0x08}
};


/** TrashMachine *********************************************/
/** Deallocate all resources taken by InitMachine().        **/
/*************************************************************/
void TrashMachine(void)
{
#ifndef N64
  if(Verbose) printf("Shutting down...\n");

  if(ColCnt) XFreeColors(Dsp,DefaultCMap,ColBuf,ColCnt,0); 

  if(SBuf) free(SBuf);

  if(Dsp&&Wnd)
  {
#ifdef MITSHM
    if(UseSHM)
    {
      XShmDetach(Dsp,&SHMInfo);
      if(SHMInfo.shmaddr) shmdt(SHMInfo.shmaddr);
      if(SHMInfo.shmid>=0) shmctl(SHMInfo.shmid,IPC_RMID,0);
    }
    else
#endif MITSHM
    if(Img) XDestroyImage(Img);
    XAutoRepeatOn(Dsp);
  }
  if(Dsp) { XAutoRepeatOn(Dsp);XCloseDisplay(Dsp); }

#ifdef SOUND
  StopSound();TrashSound();
#endif SOUND
#endif
}


/** InitMachine **********************************************/
/** Allocate resources needed by Unix/X-dependent code.     **/
/*************************************************************/
int InitMachine(void)
{
  //Screen *Scr;
  //XEvent E;
  int J;

#ifdef SOUND
  if(UseSound) InitSound();
#endif SOUND

  if(Verbose)
    printf("Initializing Unix/X drivers:\n  Opening display...");
    
  #ifndef N64
  Dsp=XOpenDisplay(NULL);
  if(!Dsp) { if(Verbose) printf("FAILED\n");return(0); }

  Scr=DefaultScreenOfDisplay(Dsp);
  White=WhitePixelOfScreen(Scr);
  Black=BlackPixelOfScreen(Scr);
  DefaultGC=DefaultGCOfScreen(Scr);
  DefaultCMap=DefaultColormapOfScreen(Scr);

  if(Verbose) printf("OK\n  Opening window...");
  Wnd=
    XCreateSimpleWindow
    (Dsp,RootWindowOfScreen(Scr),0,0,WIDTH,HEIGHT,0,White,Black);
  if(!Wnd) { if(Verbose) printf("FAILED\n");return(0); }

  {
    XSizeHints Hints;
    XWMHints WMHints;

    Hints.flags=PSize|PMinSize|PMaxSize;
    Hints.min_width=Hints.max_width=Hints.base_width=WIDTH;
    Hints.min_height=Hints.max_height=Hints.base_height=HEIGHT;
    WMHints.input=True;WMHints.flags=InputHint;
    XSetWMHints(Dsp,Wnd,&WMHints);
    XSetWMNormalHints(Dsp,Wnd,&Hints);
    XStoreName(Dsp,Wnd,Title);
  }

  XSelectInput
  (Dsp,Wnd,FocusChangeMask|ExposureMask|KeyPressMask|KeyReleaseMask);
  XMapRaised(Dsp,Wnd);
  XClearWindow(Dsp,Wnd);
  XAutoRepeatOff(Dsp);
  XWindowEvent(Dsp,Wnd,ExposureMask,&E);

  if(Verbose) printf("OK\n  Allocating sprite buffer...");
  SBuf=(byte *)malloc(sizeof(byte)*256*212);
  if(!SBuf) { if(Verbose) printf("FAILED\n");return(0); }

#ifdef MITSHM
  if(UseSHM)
  {
    if(Verbose) printf("OK\n  Using shared memory:\n    Creating image...");
    Img=
      XShmCreateImage
      (Dsp,DefaultVisualOfScreen(Scr),8,ZPixmap,NULL,&SHMInfo,WIDTH,HEIGHT);
    if(!Img) { if(Verbose) printf("FAILED\n");return(0); }

    if(Verbose) printf("OK\n    Getting SHM info...");
    SHMInfo.shmid=
      shmget(IPC_PRIVATE,Img->bytes_per_line*Img->height,IPC_CREAT|0777);
    if(SHMInfo.shmid<0) { if(Verbose) printf("FAILED\n");return(0); }

    if(Verbose) printf("OK\n    Allocating SHM...");
    XBuf=(byte *)(Img->data=SHMInfo.shmaddr=shmat(SHMInfo.shmid,0,0));
    if(!XBuf) { if(Verbose) printf("FAILED\n");return(0); }

    SHMInfo.readOnly=False;
    if(Verbose) printf("OK\n    Attaching SHM...");
    if(!XShmAttach(Dsp,&SHMInfo))
    { if(Verbose) printf("FAILED\n");return(0); }
  }
  else
#endif
  {
    if(Verbose) printf("OK\n  Allocating screen buffer...");
    XBuf=(byte *)malloc(sizeof(byte)*HEIGHT*WIDTH);
    if(!XBuf) { if(Verbose) printf("FAILED\n");return(0); }

    if(Verbose) printf("OK\n  Creating image...");
    Img=
      XCreateImage
      (Dsp,DefaultVisualOfScreen(Scr),8,ZPixmap,0,XBuf,WIDTH,HEIGHT,8,0);
    if(!Img) { if(Verbose) printf("FAILED\n");return(0); }
  }
#endif

  ColCnt=0;XPal0=Black;
  for(J=0;J<16;J++)
  { Pal[J][0]=Pal[J][1]=Pal[J][2]=0;XPal[J]=Black; }

  if(Verbose) printf("OK\n");

#ifndef N64

  if(!IntSync)
  {
    if(Verbose) printf("Initializing timer for %dHz...",IFreq);
    signal(SIGVTALRM,OnTimer);
    TimerValue.it_interval.tv_sec =
    TimerValue.it_value.tv_sec    = 0L;
    TimerValue.it_interval.tv_usec=
    TimerValue.it_value.tv_usec   = 1000000L/IFreq;
    J=setitimer(ITIMER_VIRTUAL,&TimerValue,NULL);
    if(Verbose) printf(J? "FAILED\n":"OK\n");
    if(J) return(0);
  }
  signal(SIGHUP,OnBreak);signal(SIGINT,OnBreak);
  signal(SIGQUIT,OnBreak);signal(SIGTERM,OnBreak);
#endif

  return(1);
}


/** Keyboard *************************************************/
/** Check for keyboard events, parse them, and modify MSX   **/
/** keyboard matrix.                                        **/
/*************************************************************/
void Keyboard(void)
{
#ifndef N64
  XEvent E;
  int J;

#ifdef SOUND
  FlushSound();  /* Flush sound stream on each interrupt */
#endif

  if(XCheckWindowEvent(Dsp,Wnd,KeyPressMask|KeyReleaseMask,&E))
  {
    J=XLookupKeysym((XKeyEvent *)&E,0);
    if(E.type==KeyPress)
    {
      switch(J)
      {
#ifdef DEBUG
        case XK_F11: Trace=!Trace;break;
#endif
        case XK_F12: CPURunning=0;break;
        case XK_F13: UseFont=UseFont? 0:FontBuf? 1:0;break;
      }
      J&=0xFF;KeyMap[Keys[J][0]]&=~Keys[J][1];
    }
    else { J&=0xFF;KeyMap[Keys[J][0]]|=Keys[J][1]; }
  }

  for(J=0;XCheckWindowEvent(Dsp,Wnd,FocusChangeMask,&E);)
    J=(E.type==FocusOut); 
  if(SaveCPU&&J)
  {
    XAutoRepeatOn(Dsp);   /* Turn autorepeat on       */
#ifdef SOUND
    StopSound();          /* Silence the sound        */
#endif

    while(!XCheckWindowEvent(Dsp,Wnd,FocusChangeMask,&E)&&CPURunning)
    {
      if(XCheckWindowEvent(Dsp,Wnd,ExposureMask,&E)) PutImage();
      XPeekEvent(Dsp,&E);
    }

    XAutoRepeatOff(Dsp);  /* Turn autorepeat off      */

#ifdef SOUND
    ResumeSound();        /* Switch the sound back on */
#endif
  }
#endif
}


/** InitScrF *************************************************/
/** Generic function to be called when screen is changed.   **/
/*************************************************************/
int InitScrF(void)
{
  static OldMode=0;
  int	 Color;
  register byte J,I;
  register word R,G,B;
  unsigned long L;
 
  if(Verbose&0x02) printf("Screen %d initialized\n",ScrMode);

  if(OldMode==8)
  { 
    for(R=0;R<32;R+=8)
      for(G=0;G<256;G+=64)
        for(B=0;B<4;B++)
        {
          L=BPal[R|G|B];
          // if(L!=Black) XFreeColors(Dsp,DefaultCMap,&L,1,0);
        }
  }

  if(ScrMode==8)
  {  
    for(R=0;R<4;R++)
      for(G=0;G<4;G++)
        for(B=0;B<4;B++)
        {
          /*
          Color.flags=DoRed|DoGreen|DoBlue;
          Color.red=(word)R<<14;
          Color.green=(word)G<<14;
          Color.blue=(word)B<<14;
          J=XAllocColor(Dsp,DefaultCMap,&Color)? Color.pixel:Black;
          */
          I=B|(R<<3)|(G<<6);
          BPal[I]=BPal[I|0x04]=BPal[I|0x20]=BPal[I|0x24]=J;
        }  
  }

  OldMode=ScrMode;return(1);
}


/** ColorScrF ************************************************/
/** Generic function to be called when colors are changed.  **/
/*************************************************************/
void ColorsScrF(void)
{
  int Color;
  int N;
  byte J;

  XPal[0]=XPal0;

  for(J=0;(J<16)&&(ColCnt<16*256);J++)
    if(
        (Palette[J][0]!=Pal[J][0])||
        (Palette[J][1]!=Pal[J][1])||
        (Palette[J][2]!=Pal[J][2])
      ) if(!Palette[J][0]&&!Palette[J][1]&&!Palette[J][2])
        {
          if(XPal[J]!=Black) ColBuf[ColCnt++]=XPal[J];
          Pal[J][0]=Pal[J][1]=Pal[J][2]=0;
          XPal[J]=Black;
        }
        else
        { /*
          Color.flags=DoRed|DoGreen|DoBlue;
          Color.red=(word)Palette[J][0]<<8;
          Color.green=(word)Palette[J][1]<<8;
          Color.blue=(word)Palette[J][2]<<8;
          if(XAllocColor(Dsp,DefaultCMap,&Color))
          {
            if(XPal[J]!=Black) ColBuf[ColCnt++]=XPal[J];
            Pal[J][0]=Palette[J][0];
            Pal[J][1]=Palette[J][1];
            Pal[J][2]=Palette[J][2];
            XPal[J]=Color.pixel;
          }
          */
        }

  XPal0=XPal[0];
  XPal[0]=(!BGColor||SolidColor0)? XPal0:XPal[BGColor];

  N=ScanLine? WIDTH*(ScanLine+(HEIGHT-192)/2):0;
  J=(ScrMode&&(ScrMode!=MAXSCREEN+1))? 0:BGColor;
  memset(XBuf+N,XPal[J],WIDTH*HEIGHT-N);
}


/** Part of the code common for Unix/X and MSDOS drivers *****/ 
#include "Common.h"


/** RefreshScr6 **********************************************/
/** Function to be called to update SCREEN 6.               **/
/*************************************************************/
void RefreshScr6(register byte Y1,register byte Y2)
{
  register byte X,J,C;
  register byte *P,*T,*R;

  P=XBuf+WIDTH*(HEIGHT-192)/2+Y1*WIDTH;

  if(ScreenON)
  {
    ColorSprites();

    R=SBuf+((word)Y1<<8);
    P+=(WIDTH-256)/2;
    J=VDP[23]+Y1;
    T=ChrTab+((word)J<<7);

    for(;Y1<=Y2;Y1++,J++,P+=WIDTH-256)
    {
      if(!J) T=ChrTab;
      for(X=0;X<32;X++)
      {
        C=R[0];C=((C>>4)|C)&0x0F;P[0]=XPal[C? C:T[0]>>6];
        C=R[1];C=((C>>4)|C)&0x0F;P[1]=XPal[C? C:(T[0]>>2)&0x03];
        C=R[2];C=((C>>4)|C)&0x0F;P[2]=XPal[C? C:T[1]>>6];
        C=R[3];C=((C>>4)|C)&0x0F;P[3]=XPal[C? C:(T[1]>>2)&0x03];
        C=R[4];C=((C>>4)|C)&0x0F;P[4]=XPal[C? C:T[2]>>6];
        C=R[5];C=((C>>4)|C)&0x0F;P[5]=XPal[C? C:(T[2]>>2)&0x03];
        C=R[6];C=((C>>4)|C)&0x0F;P[6]=XPal[C? C:T[3]>>6];
        C=R[7];C=((C>>4)|C)&0x0F;P[7]=XPal[C? C:(T[3]>>2)&0x03];
        R+=8;P+=8;T+=4;
      }
    }  
  }
  else memset(P,XPal[BGColor],(Y2-Y1+1)*WIDTH);

  if(EndOfFrame) PutImage();
}
  

/** RefreshScr7 **********************************************/
/** Function to be called to update SCREEN 7.               **/
/*************************************************************/
void RefreshScr7(register byte Y1,register byte Y2)
{
  register byte X,J,C;
  register byte *P,*T,*R;

  P=XBuf+WIDTH*(HEIGHT-192)/2+Y1*WIDTH;

  if(ScreenON)
  {
    ColorSprites();

    R=SBuf+((word)Y1<<8);
    P+=(WIDTH-256)/2;
    J=VDP[23]+Y1;
    T=ChrTab+((word)J<<8);

    for(;Y1<=Y2;Y1++,J++,P+=WIDTH-256)
    {
      if(!J) T=ChrTab;
      for(X=0;X<32;X++)
      {
        C=R[0];C=((C>>4)|C)&0x0F;P[0]=XPal[C? C:T[0]>>4];
        C=R[1];C=((C>>4)|C)&0x0F;P[1]=XPal[C? C:T[1]>>4];
        C=R[2];C=((C>>4)|C)&0x0F;P[2]=XPal[C? C:T[2]>>4];
        C=R[3];C=((C>>4)|C)&0x0F;P[3]=XPal[C? C:T[3]>>4];
        C=R[4];C=((C>>4)|C)&0x0F;P[4]=XPal[C? C:T[4]>>4];
        C=R[5];C=((C>>4)|C)&0x0F;P[5]=XPal[C? C:T[5]>>4];
        C=R[6];C=((C>>4)|C)&0x0F;P[6]=XPal[C? C:T[6]>>4];
        C=R[7];C=((C>>4)|C)&0x0F;P[7]=XPal[C? C:T[7]>>4];
        R+=8;P+=8;T+=8;
      }
    }
  }
  else memset(P,XPal[BGColor],(Y2-Y1+1)*WIDTH);

  if(EndOfFrame) PutImage();
}


/** RefreshScr8 **********************************************/
/** Function to be called to update SCREEN 8.               **/
/*************************************************************/
void RefreshScr8(register byte Y1,register byte Y2)
{
  static byte SprToScr[16] =
  {
    0x00,0x02,0x10,0x12,0x80,0x82,0x90,0x92,
    0x49,0x4B,0x59,0x5B,0xC9,0xCB,0xD9,0xDB
  };
  register byte *P,*T,*R;
  register byte J,C;
  register word X;
  
  P=XBuf+WIDTH*(HEIGHT-192)/2+Y1*WIDTH;

  if(ScreenON)
  {
    ColorSprites();

    R=SBuf+((word)Y1<<8);
    P+=(WIDTH-256)/2;
    J=VDP[23]+Y1;
    T=ChrTab+((word)J<<8);

    for(;Y1<=Y2;Y1++,J++,P+=WIDTH-256)
    {
      if(!J) T=ChrTab;
      for(X=0;X<256;X++,T++,R++)
      { C=((*R>>4)|*R)&0x0F;*P++=BPal[C? SprToScr[C]:*T]; }
    }
  }
  else memset(P,XPal[BGColor],(Y2-Y1+1)*WIDTH);

  if(EndOfFrame) PutImage();
}


/** RefreshTx80 **********************************************/
/** Function to be called to update TEXT80.                 **/
/*************************************************************/
void RefreshTx80(register byte Y1,register byte Y2)
{
  register byte X,Y,J,K,M,FC,BC;
  register byte *P,*S,*C,*T;
  byte *G;

  P=XBuf+WIDTH*(HEIGHT-192)/2;
  G=UseFont&&FontBuf? FontBuf:ChrGen;

  if(ScreenON)
  {
    P+=(WIDTH-320)/2;C=ColTab;T=ChrTab;M=0x80;

    for(Y=ScanLines212? 26:24;Y;Y--,P+=WIDTH*8-4*80)
      for(X=0;X<80;X++,T++)
      {
        S=G+((long)*T<<3);
        if(*C&M) { FC=XPal[XFGColor];BC=XPal[XBGColor]; }
        else     { FC=XPal[FGColor];BC=XPal[BGColor]; }
        for(J=0;J<8;J++)
        {
          K=*S++;
          P[0]=K&0xC0? FC:BC;
          P[1]=K&0x30? FC:BC;
          P[2]=K&0x08? FC:BC;
          P[3]=K&0x04? FC:BC;
          P[4]=BC;P+=WIDTH;
        }
        P+=4-WIDTH*8;
        if(!(M>>=1)) { M=0x80;C++; } 
      }
  }
  else memset(P+(word)Y1*WIDTH,XPal[BGColor],(Y2-Y1+1)*WIDTH);
    
  if(EndOfFrame) PutImage();
}
