/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                          fMSX.c                         **/
/**         Dreamcast version 0.2b By Jos Kwanten           **/
/**                                                         **/
/*************************************************************/

#include "dream.h"
#include "MSX.h"
#include "mem.h"

int configure_game(char *rom_name_a, 
                       int  *rom_size_a,
                       char *rom_name_b, 
                       int  *rom_size_b,
                       int  *iperiod,
                       byte *frame_skip,
                       byte *rom_type_a,
                       byte *rom_type_b,
                       char *dsk_name_a);

int get_Option(char options[256][15],
               int numOptions,
               int ypos,
               uint8 mcont,
               int defaultOption);

int atoi (const char *str);

char names[256][15];
char dsks[256][15]; // Names of the disk images
char iperiods[10][15] = {"10000\0", "20000\0", "30000\0", "40000\0", "50000\0", 
												"59667\0", "70000\0", "80000\0", "90000\0", "100000\0"};
char uperiods[6][15] = {"None\0", "1\0", "2\0", "3\0", "4\0", "5\0"};
char rom_types[7][15] =     {"KONAMI/8kB\0","MSXDOS2\0","KONAMI5/8kB\0",
                          "KONAMI4/8kB\0","ASCII/8kB\0","ASCII/16kB\0",
                          "GAME MASTER 2\0"};

int  sizes[256];


/* include this file at last to align it on a four-byte memory address */
#include "arm/armcode.h"

void clearscr(void)
{
	int x, y;
	for(x = 0; x < 640; x++)
	    for(y = 0; y < 480; y++)
          vram_s[640*y + x] = 0;	        	
}


int load_system_roms(void)
{
	DIR *d;
	int amt, fd;
	d = cd_opendir("/");
  fd = cd_open("MSX2.ROM", O_RDONLY);
	if (fd < 0) {
		printf("Can't open file /MSX2.ROM for reading.\r\n");
		return 0;
	}
	
	amt = cd_read(fd, msxrom, 32768);
	cd_close(fd);
  
  fd = cd_open("MSX2EXT.ROM", O_RDONLY);
	if (fd < 0) {
		printf("Can't open file /MSX2.ROM for reading.\r\n");
		return 0;
	}
	
	amt = cd_read(fd, msx2extrom, 16384);
	cd_close(fd);

  fd = cd_open("DISK.ROM", O_RDONLY);
  if (fd < 0) {
    printf("Can't open /DISK.ROM for reading.\r\n");
  }
  else
  {
    amt = cd_read(fd, diskrom, 0x4000);
    cd_close(fd);
  }

	cd_closedir(d);
	return 1;
}      

int dc_main(void)
{
	char rom_name[255];
	char rom_name_b[255];
	int size;
	int size_b;
	int y1;
	serial_disable();
//	serial_init(57600);
	dc_setup(DM_640x480, PM_RGB565);
	
	if(load_system_roms()<= 0)
	  {
	    y1 = 25; 
      bfont_draw_str(vram_s+y1*640+10, 640, "System roms missing...");
	    return(0);
	  }
	
  while(1)
  {
  
  dc_setup(DM_640x480, PM_RGB565);
  
  clearscr();
	

  DiskA[0] = '\0';
	if(configure_game(rom_name, &size, 
	                      rom_name_b, 
	                      &size_b, 
	                      &IPeriod, 
	                      &UPeriod,
	                      &ROMTypeA,
                        &ROMTypeB,
                        DiskA) < 0)
	  {
	    return 0;
	  }
	
#ifdef BPP24
	dc_setup(DM_320x240, PM_RGB888);
#else
//  dc_setup(DM_640x480, PM_RGB565);
	dc_setup(DM_320x240, PM_RGB555);
#endif
	
  init_mem();
  /* Start a song playing */
	snd_init();
	
	/* Switch channels to mono if uncommented */
	snd_load_arm(armcode, sizeof(armcode));
  
  TrapBadOps = 0;
  RAMPages = 8;
  VRAMPages = 8;
  MSXVersion = 1;
  IntSync  = 1;
  Verbose  = 0;
  if(!InitMachine()) return(1);
  StartMSX(rom_name, size, rom_name_b, size_b);
  TrashMSX();
  TrashMachine();
  }
  return(0);
}

/* my_strcpy */
void my_strcpy(char *to, char *from) 
{
 while(*from)
   {
     *to++ = *from++;
   }
   
 *to = 0;  
}


int get_Option(char options[256][15],
               int numOptions,
               int ypos,
               uint8 mcont,
               int defaultOption)
{
	int i, j, k, max;
  int y1;
	cont_cond_t cond;

	if (cont_get_cond(mcont, &cond)) 
	{
		printf("Error getting controller status first\r\n");
		return -1;
	}

  max = numOptions - 1;
  y1 = ypos;
  i = defaultOption;

	while(cond.buttons & CONT_START)
	{
    // Stick functions
    if(!(cond.buttons  & CONT_DPAD_UP))
    {
      i ++;
      if(i > max)
        i = 0;
  
      for(j = 0; j < 24; j++)
      {
        for(k = 0; k < 640; k++)
          vram_s[(y1+j)*640+k] = 0;
      } 
      bfont_draw_str(vram_s+y1*640+10+5*12, 640, options[i]);
    }

     // Stick functions
    if(!(cond.buttons  & CONT_DPAD_DOWN))
    {
      i --;
      if(i < 0)
        i = max;
      for(j = 0; j < 24; j++)
      {
        for(k = 0; k < 640; k++)
          vram_s[(y1+j)*640+k] = 0;
      }       

      bfont_draw_str(vram_s+y1*640+10+5*12, 640, options[i]); 	  
    }

    if(!(cond.buttons  & CONT_X))
    {
      printf("Escape from DreamMSX\r\n");
      return -1;
    }

    sleep(100);

    /* Check for start on the controller */
    if(cont_get_cond(mcont, &cond)) 
    {
      printf("Error getting controller status menu\r\n");
      return -1;
    }
  }
  return i;
}

int configure_game(char *rom_name_a, 
                   int  *rom_size_a,
                   char *rom_name_b, 
                   int  *rom_size_b,
                   int  *iperiod,
                   byte *frame_skip,
                   byte *rom_type_a,
                   byte *rom_type_b,
                   char *dsk_name_a)
{
	int y1;
	DIR *d;
	struct dirent *de;
	int i, max, j;
	/* gamee names */
	uint8 mcont;


	/* Read and print the root directory */
  i = 0;
  my_strcpy(names[i], "NONE");
  sizes[i] = 0;
  i++;

//  d = cd_opendir("/roms");
  d = cd_opendir("/");
	if (!d) 
	{
		printf("Can't open rom folder!\r\n");
	}
  else
  {
	  while ( (de=cd_readdir(d)) ) 
	  {
	  	if(de->d_size > 0)
	  	  {
          my_strcpy(names[i], de->d_name);  		
		      //printf("%s  /  ", names[i]);
		      sizes[i] = de->d_size;
			    //printf("%d\r\n", de->d_size);
			    i++;
			  }
		}
	  cd_closedir(d);
  }

  j = 0;
  my_strcpy(dsks[j], "NONE");
  j++;

  d = cd_opendir("/");
  if (!d)
  {
		printf("Can't open dsk folder!\r\n");
  }
  else
  {
	  while ( (de=cd_readdir(d)) ) 
	  {
	  	if(de->d_size > 0)
	  	  {
          my_strcpy(dsks[j], de->d_name);  		
		      //printf("%s\r\n", dsks[j]);
			    j++;
			  }
		}
	  cd_closedir(d);
  }

  max = i;
	
	y1 = 25;
	bfont_draw_str(vram_s+y1*640+10, 640, "Game rom 1");
	y1 += 25;
	bfont_draw_str(vram_s+y1*640+10+5*12, 640, names[0]);

	y1 += 25;
	bfont_draw_str(vram_s+y1*640+10, 640, "Game rom 2");
	y1 += 25;
	bfont_draw_str(vram_s+y1*640+10+5*12, 640, names[0]);


	y1 += 25;
	bfont_draw_str(vram_s+y1*640+10, 640, "I_Period");
	y1 += 25;
	bfont_draw_str(vram_s+y1*640+10+5*12, 640, iperiods[5]);
	
	y1 += 25;
	bfont_draw_str(vram_s+y1*640+10, 640, "Frame skip");
	y1 += 25;
	bfont_draw_str(vram_s+y1*640+10+5*12, 640, uperiods[0]);
	y1 += 25;
	bfont_draw_str(vram_s+y1*640+10, 640, "ROM 1 Type");
	y1 += 25;
	bfont_draw_str(vram_s+y1*640+10+5*12, 640, rom_types[0]);
	y1 += 25;
	bfont_draw_str(vram_s+y1*640+10, 640, "ROM 2 Type");
	y1 += 25;
	bfont_draw_str(vram_s+y1*640+10+5*12, 640, rom_types[0]);
	y1 += 25;
	bfont_draw_str(vram_s+y1*640+10, 640, "Disk 1");
	y1 += 25;
	bfont_draw_str(vram_s+y1*640+10+5*12, 640, dsks[0]);
	y1 = 180;
	bfont_draw_str(vram_s+y1*640+300, 640, "Marco Aurelio Edition");

  y1 = 50;
	
	mcont = maple_controller_addr();
	if (!mcont) {
		printf("No controllers attached\r\n");
		return -1;
	}
	
  i = get_Option(names, max, y1, mcont, 0);
  if (i == -1)
    return i;
  my_strcpy(rom_name_a, names[i]);
  *rom_size_a = sizes[i];
  printf("Game name = %s, size = %d \r\n", rom_name_a, *rom_size_a);

  sleep(100);
  y1 += 50;
  i = get_Option(names, max, y1, mcont, 0);
  if (i == -1)
    return i;
  my_strcpy(rom_name_b, names[i]);
  *rom_size_b = sizes[i];
  printf("Game name = %s, size = %d \r\n", rom_name_b, *rom_size_b);
  
  /* configure I_PERIOD */
  sleep(100);
  y1 += 50;
  max = 10;
  i = 5;
  i = get_Option(iperiods, max, y1, mcont, i);
  if (i == -1)
    return i;
	*iperiod = atoi(iperiods[i]);
  printf("IPeriod = %d\r\n", *iperiod);
  
  /* configure U_PERIOD */
  sleep(100);
  y1 += 50;
  max = 6;
  i = 0;
  i = get_Option(uperiods, max, y1, mcont, i);
  if (i == -1)
    return i;
  *frame_skip = (byte) i + 1;
  printf("UPeriod = %d\r\n", *frame_skip);
  

/* configure rom_type */
  sleep(100);
  y1 += 50;
  max = 7;
  i = 0;
  i = get_Option(rom_types, max, y1, mcont, i);
  if (i == -1)
    return i;
  *rom_type_a = (byte) i; 
  printf("rom_type_a = %s\r\n", rom_types[*rom_type_a]);

  sleep(100);
  y1 += 50;
  max = 7;
  i = 0;
  i = get_Option(rom_types, max, y1, mcont, i);
  if (i == -1)
    return i;
  *rom_type_b = (byte) i; 
  printf("rom_type_b = %s\r\n", rom_types[*rom_type_b]);

/* Get disk 1 */

  sleep(100);
  y1 += 50;
  max = j;
  i = get_Option(dsks, max, y1, mcont, 0);
  if (i == -1)
    return i;
  if (i == 0)
    dsk_name_a[0] = '\0';
  else
    my_strcpy(dsk_name_a, dsks[i]);
  printf("Option selected is %d, Disk 1 is = %s\r\n", i, dsk_name_a);
  return 0;	
}



int atoi(const char* str)
{
	int length;
	int value;
	int pow;

	value = 0;
	pow = 1;
	for (length=0; str[length] != '\0'; length++);

	while (length > 0)
		{
			length--;
			value += (pow * (str[length] - '0'));
			pow *= 10;
		}
	return value;
}
