#define N64

#ifndef N64
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#endif

#ifdef N64
#define  NULL 0
#include <mmio.h>
#endif

typedef unsigned char byte;

int InitSound(void);
void TrashSound(void);
void FlushSound(void);
void PSGOut(byte R,byte V);
void SCCOut(byte R,byte V);

unsigned int Verbose = 0;     /* Verboseness level 0/1      */
unsigned int IPeriod = 0;     /* Microseconds per interrupt */

char *Options[] = { "freq",NULL };

int main(int argc,char *argv[])
{
  FILE *F;
  char S[32];
  int I,J,R,V;
  long I,L;

  if(argc<2)
  {
    puts("PSG Player by Marat Fayzullin   v1.0 (C)FMS 1996");
    puts("Usage: psgplay [-freq <20..500>] <file1.psg> <file2.psg> ...");
    return(1);
  }

  if(!InitSound())
  { printf("%s: Sound initialization failed!\n",argv[0]);return(1); }

  for(J=1;J<argc;J++)
    if(argv[J][0]=='-')
    {
      for(I=0;Options[I];I++)
        if(!strcmp(&argv[J][1],Options[I]))
        {
          switch(I)
          {
            case 0:  J++;
                     if(J<argc) L=atol(argv[J]);
                     else printf("%s: No frequency supplied\n",argv[0]);
                     if((L>=20)||(L<=500))
                       IPeriod=(unsigned int)(1000000L/L);
                     else
                       printf("%s: Frequency must be in 20-500Hz range\n",argv[0]);
                     break;
            default: printf("%s: Unknown option '%s'\n",argv[0],argv[J]);
          }
          break;
        }
    }
    else
      if(F=fopen(argv[J],"rb"))
      {
        if(fread(S,1,16,F)!=16) I=0;
        else I=!strncmp(S,"PSG\032",4);

        if(!I) printf("%s: '%s' is not a .PSG file\n",argv[0],argv[J]);
        else
        {
          if(!IPeriod)
            IPeriod=S[5]? 1000000L/((unsigned char *)S)[5]:5000;

          while((R=fgetc(F))>=0)
            switch(R)
            {
              case 0xFF: FlushSound();usleep(IPeriod);break;
              case 0xFE: FlushSound();usleep(IPeriod*fgetc(F));break;
              default:   V=fgetc(F);
                         if(R<14) PSGOut(R,V);
                         if(R>15) SCCOut(R-16,V);
            }
        }
        fclose(F);
      }

  TrashSound();
}
